#!/usr/bin/perl -w

$| = 1;

# This is a Perl program which reads a GNU ddrescue 1.0 "logfile" (which is
# really a sector list file), and generates a .bat file of SpinRite 6.0
# commands to try to recover the unrecovered regions.  After SpinRite has
# completed its work, you can run ddrescue 1.0 again to add the recovered
# data to the recovery image.
#
# This program can be run under any system that supports Perl 4 or 5
# (Linux, DOS, Windows, etc.).
#
# Note #1:  ddrescue runs under Linux, and SpinRite runs under MS-DOS or
# FreeDOS.  So (obviously) some rebooting will be required.
#
# Note #2:  You must use Antonio Diaz's GNU ddrescue 1.0 or later.  You
# cannot use ddrescue 0.9 or earlier, nor can you use Garloff's dd_rescue
# (which is an entirely different program).
#
# Note #3:  To use this program you MIGHT need to modify it!
# The part that you might need to modify is between the "=========" divider
# comment lines.  These are the changes you might need to make:
#
#   1) This program expects to find the output of Linux 'fdisk -lu {device}'
#      for the drive in a file named 'fdisk-lu_out.txt'.  If that's not
#      right, then rename it or edit $fdisk_out.
#
#   2) If your ddrescue sector list file is not named drive.log, then
#      rename it, or else adjust $ddrescuelog to read the right file.
#
#   3) If the deduced "$drive" does not match how SpinRite numbers your
#      disk drives, then override it by changing the assignment statement
#      to something other than -1.
#      (Use "spinrite list" -- I don't think there is any good way to more
#      reliably automate this.)
#
#   4) Run SpinRite and verify that it numbered the partitions the same
#      way that you numbered them in the starting[]/ending[] arrays which
#      contain this program's representation of the partition table.
#      (The SpinRite "item number" is the index to the starting[] and
#      ending[] arrays.)
#
# Then, to use this program, simply run it, as in this example:
#
#    perl ddr2sr.pl
#
# It will create an output file called SR6CMDS.BAT which you'll probably
# want to copy to your SpinRite 6.0 diskette:
#
#    cp SR6CMDS.BAT /mnt/floppy/
#
# (where /mnt/floppy/ is the diskette drive device name under your version
# of Linux, containing your SpinRite 6.0 diskette)
#
# Save a copy of the ddrescue .log file at this point, so you have a
# record of which sectors are questionable:
#
#    cp drive.log /mnt/floppy/
#
# Then run SpinRite and adjust its options according to your preferences.
# Be sure to tell it to create a log file, since that will record which
# sectors were not completely recoverered.  Also, I recommend "short phrases,"
# and that you disable benchmark reporting.  If your disk image contains a
# FAT32 partition, then be sure to run SpinRite under the "DOS 7" that is
# part of Win98 or WinMe or Win95OSR2, so that SpinRite can correctly
# determine which files are damaged.
#
# Then run the generated SR6CMDS.BAT file to make SpinRite try to recover the
# desired sectors.
#
# Then restart Linux, and run ddrescue again, to add the recovered
# sectors to your recovered disk image.
#
# Note: SpinRite's log file is your primary record of which sectors were
# unrecoverable.  The saved ddrescue log file (from before you ran
# SpinRite) is your record of which sectors were unrecovered by ddrescue
# before you ran SpinRite.
#
#
# REFERENCES:
#
#   GNU ddrescue 1.0 (free!):
#     http://www.gnu.org/software/ddrescue/ddrescue.html
#     http://savannah.gnu.org/download/ddrescue/
#     http://osx.freshmeat.net/projects/addrescue/
#
#   SpinRite 6.0 ($89.00, and worth every penny):
#     http://www.grc.com/sr/spinrite.htm
#
#   Perl 5 (free, but big):
#     http://www.perl.org/
#     http://www.cpan.org/
#     http://www.perl.com/
#
#   Perl 4 (free and tiny):
#     http://www.burtonsys.com/auxil/#tinyperls
#     (As small as 138K, so it fits on a DOS diskette.)
#
#   This program (free):
#     http://www.burtonsys.com/download/ddr2sr.zip
#
#   Tool to convert Linux/Unix End-Lines (CR) to DOS/Windows (CR+LF):
#     http://www.burtonsys.com/download/fixtext.zip
#     (You might need this to view ddrescue's logfile / sector list file.)
#
#
# This program is UNCOPYRIGHTED.  It was written by:
#
#    Dave Burton
#    Burton Systems Software
#    Tel: +1 919-481-0149
#    Email: dave at burtonsys.com
#
#
# v.1 - 20-Jul-2005 - first version.  -DAB
# v.2 - 24-Sep-2005 - second version, fixes bugs & adds region merge.  -DAB
# v.3 - 17-Mar-2006 - third version, parses 'fdisk -lu' outptut.  -DAB

#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "13, 13-Sep-06";
# the TLIB Version Control version number and last-modified date


# ============================================================================
$drive = -1;  # "-1" for best guess, or specify a non-negative integer matching how SpinRite numbers the drives
$ddrescuelog = 'drive.log';  # sector list file
$fdisk_out = 'fdisk-lu_out.txt';  # output of 'fdisk -lu' command
$verbose = 1;
# ============================================================================


open( FDISK_OUT, "<$fdisk_out") || die "ERROR: Could not read '$fdisk_out', $!\n";

# Read FDISK_OUT, to get the partition table information.  Here's an example
# (note the blank line at the beginning):
#----------------------------------------------------------------------
#
# Disk /dev/hdd: 40.0 GB, 40000000000 bytes
# 255 heads, 63 sectors/track, 4863 cylinders, total 78125000 sectors
# Units = sectors of 1 * 512 = 512 bytes
#
#    Device Boot      Start         End      Blocks   Id  System
# /dev/hdd1              63       80324       40131   de  Dell Utility
# /dev/hdd2   *       80325    78108029    39013852+   7  HPFS/NTFS
#----------------------------------------------------------------------
#
# Here's an example of the result:
# This is how the drive is partitioned, all numbers are LBA sector numbers:
# 0 - 62 MBR
# 63 - 80,324 in a Dell Utility partition
# 80,325 - 78,108,029 in one NTFS partition
#
# These are indexed by SpinRite "Item" number (partition number, sorta):
# $starting[0] = 0;
# $ending[0] = 62;
# $starting[1] = 63;
# $ending[1] = 80324;
# $starting[2] = 80325;
# $ending[2] = 78108029;

$saw_Units_line = $num_partitions = 0;
while (<FDISK_OUT>) {
   $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
   if ($_ eq 'Units = sectors of 1 * 512 = 512 bytes') {
      $saw_Units_line = 1;
   }
   # match a line like this:
   # '/dev/hdg3       305154675   312496379     3670852+  db  CP/M / CTOS / ...'
   if ($_ =~ /^(\/dev\/\w*[a-z])(\d+)\s+(\*|)\s*(\d+)\s+(\d+)\s+(\d+)(\+|)\s+(\w{1,2})\s*(.*)$/) {
      ($dev, $partnum, $bootind, $start, $end, $blocks, $halfblockind, $id, $system) =
        ($1, $2, $3, $4, $5, $6, $7, $8, $9);
      if ($verbose) {
         print "dbg: device= $dev$partnum\n";
         print "     bootind=$bootind\n";
         print "     start=$start\n";
         print "     end=$end\n";
         print "     blocks=$blocks";
         if ('' ne $halfblockind) {
            print ".5";
         }
         print "\n     id=$id\n";
         print "     system=$system\n";
      }
      # if ($system =~ /(NTFS|FAT)/) {
      #    $driveletter = pack( 'c', $num_driveletters+ord('C') );
      #    $num_driveletters++;
      # } else {
      #    $driveletter = '?';
      # }
      $start += 0;  # convert to integer;
      if ((0 == $num_partitions) && ($start > 0)) {
         # typically the first $start is 63, so insert an array entry for the MBR, in sectors 0-62:
         $starting[$num_partitions] = 0;
         $ending[$num_partitions] = ($start-1);
         $num_partitions = 1;
      }
      $starting[$num_partitions] = $start;
      $ending[$num_partitions] = 0+$end;
      # $driveletters[$num_partitions] = $driveletter;
      # Use first /dev/hdx line to deduce a best guess for the drive number,
      # unless it was already assigned.
      if ((-1 == $drive) && ($dev =~ /([a-z])$/)) {
         $device_letter = $1;
         $drive = ord($device_letter) - ord('a');  # 0 for /dev/hda, 3 for /dev/hdd
         print "Note: SpinRite drive number deduced to be '$drive' -- better double check it!\n" .
               "If that's wrong, then edit '$0' and change the '\$drive = -1;' line.\n" .
               "(Use 'SPINRITE LIST' under DOS to find out what it should be.)\n";
      }
      $num_partitions++;
   }
}

# Result is the partition table, represented by @starting and @ending
if (!$saw_Units_line) {
   print "ERROR: '$fdisk_out' does not look like the output of \"fdisk -lu\".\n";
}
if (!$num_partitions) {
   print "ERROR: no recognizable partition table lines seen in '$fdisk_out'.\n";
}


# ============================================================================
# If necessary, you could insert overrides/corrections to the $starting/$ending table here.
# ============================================================================


print "\n";
print "\$drive = $drive;  # This should match the SpinRite drive number\n";
for ($i = 0; $i <= $#starting; $i++) {
   print "\$starting[$i] = " . $starting[$i] . ";  \$ending[$i] = " . $ending[$i] . ";  # Spinrite Item no. $i\n";
}
print "\n";


$ver_ordinal_and_date .= ''; # shut up a Perl warning

# Given the sector number, return the "SpinRite Item Number" (partition),
# or -1 for failure.  (It just looks it up in the starting[]/ending[] table.)
sub item_from_sect {
   local($sect) = @_;
   local($result) = -1;
   local($i);
   # print "REM dbg: item_from_sect($sect), ";
   $result = -1;
   for ($i = 0; $i <= $#starting; $i++) {
      # print "  i=$i ";
      if (($sect >= $starting[$i]) && ($sect <= $ending[$i])) {
         # printf "= result (start=%d <= sect=%d <= end=%d)", $starting[$i], $sect, $ending[$i];
         $result = $i;
         last;
      }
      # printf "(start=%d, end=%d) ", $starting[$i], $ending[$i];
   }
   # print "\n";
   return $result;
}


$perlver = "3 or earlier";
if ($] =~ /\$\$Revision\:\s*([0-9.]+)\s/) {
   $perlver = $1;  # probably 4.something
} elsif ($] =~ /([0-9][0-9.]*)/) {
   $perlver = $1;  # probably 5.something or 6.something
}
if ($verbose) {
   print "You are using Perl version $perlver\n";
}

open(INPF, "<$ddrescuelog") || die "ERROR: could not open '$ddrescuelog', $!\n";
open(SR6CMDS, ">SR6CMDS.BAT") || die "ERROR: could not create 'SR6CMDS.BAT', $!\n";

binmode SR6CMDS;  # write output in binary mode, so that we can force CR+LF line termination under Linux

$numlines = 0;
$inpline = '';


# For counting bad sectors, by doit().  When doit() calls emitit(), the
# bad sectors being processed are from $starting_bad_sector_cnt to
# $ending_bad_sector_cnt-1:
$starting_bad_sector_cnt = $ending_bad_sector_cnt = $number_of_bad_sectors = 0;

# emitit() maintains and emit_prev() uses these versions:
$prev_starting_bad_sector_cnt = $prev_ending_bad_sector_cnt = 0;


# Print a SpinRite command to recover data from a specified HDD region
sub emitit1 {
   local( $item, $startpercent, $endpercent ) = @_;
   print SR6CMDS "spinrite NoBlanking Auto Exit Level 2 Drive $drive Item $item $startpercent $endpercent" . "%%\r\n";
   # Note the doubled percentage signs at the end, because this is going into
   # a .bat file, and to use a percent sign in a bat file you must double it.
}

# these variables store the previously buffered but not-yet-printed SpinRite command, or -1 if none
$prev_item = -1;
$prev_startpercent = "-1";
$prev_endpercent = "-1";

# emit previously buffered SpinRite command
sub emit_prev {
  if ($prev_item >= 0) {
     if  (($prev_ending_bad_sector_cnt-1) > $prev_starting_bad_sector_cnt) {
        print SR6CMDS "rem - Fix bad sectors $prev_starting_bad_sector_cnt-" . ($prev_ending_bad_sector_cnt-1) . ":\r\n";
     } else {
        print SR6CMDS "rem - Fix bad sector $prev_starting_bad_sector_cnt:\r\n";
     }
     &emitit1( $prev_item , $prev_startpercent, $prev_endpercent );
     $prev_item = -1;
     $prev_startpercent = "-1";
     $prev_endpercent = "-1";
     $prev_starting_bad_sector_cnt = $prev_ending_bad_sector_cnt = 0;
  }
}

# Buffer, merge+buffer, or print a SpinRite command to recover data from a specified HDD region
sub emitit {
   local( $item, $startpercent, $endpercent ) = @_;
   if (($prev_item != $item) || (($startpercent+0.0) > ($prev_endpercent+0.0001))) {
      # print previous one
      &emit_prev();
      # buffer this one
      $prev_item = $item;
      $prev_startpercent = $startpercent;
      $prev_endpercent = $endpercent;
      $prev_starting_bad_sector_cnt = $starting_bad_sector_cnt;
      $prev_ending_bad_sector_cnt = $ending_bad_sector_cnt;
   } else {
      # merge this one with previous one
      $prev_endpercent = $endpercent;
      $prev_ending_bad_sector_cnt = $ending_bad_sector_cnt;
   }
}




######################################################################
# Here's a sample of what a ddrescue log file looks like:
#
#    # Rescue Logfile. Created by GNU ddrescue version 1.0
#    #      pos        size  status
#    0x00000000  0x0004BE00  +
#    0x0004BE00  0x00000200  -
#    0x0004C000  0x33078E00  +
#    0x330C4E00  0x00001000  -
#    0x330C5E00  0x8ACDF200  +
#
######################################################################




# This is the main loop, which reads the ddrescue log (sector list) file, and
# writes a .bat file of SpinRite commands to stdout.
sub doit {
  $starting_bad_sector_cnt = $number_of_bad_sectors = 0;
  while (<INPF>) {
     $numlines++;
     $inpline = $_;
     if ($_ =~ /^\#/) {
        next;  # skip comment line
     }
     # Parse a ddrescue log file line:
     # $_ =~ s/\r\n/\n/; -- unnecessary
     ($pos_byt, $len_byt, $status) = split( /\s+/, $_ );

     # That was easy!  Now convert byte offsets and counts to LBA sector number & sector count:
     if ($pos_byt =~ /^0x/) {
        # tricky divide by 512: drop the last 2 hexidecimal digits, and divide by 2 (which is more robust for very large integers)
        $pos_sect = substr($pos_byt,0,length($pos_byt)-2);
        $pos_sect = eval($pos_sect) / 2;
     } else {
        $pos_sect = $pos_byt / 512;
     }
     if ($len_byt =~ /^0x/) {
        $len_sect = substr($len_byt,0,length($len_byt)-2);
        $len_sect = eval($len_sect) / 2;
     } else {
        $len_sect = $len_byt / 512;
     }
     print SR6CMDS "REM sect no. = $pos_sect, count = $len_sect, '$status'\r\n";

     # The only lines we care about are the '-' lines, which indicate bad sectors
     if ('-' ne $status) {
        next;
     }

     $starting_bad_sector_cnt = $number_of_bad_sectors;
     $number_of_bad_sectors = $ending_bad_sector_cnt = $number_of_bad_sectors + $len_sect;

     # deduce the "item number" (partition number), from the LBA sector number:
     $item = &item_from_sect( $pos_sect );
     if ($item < 0) {
        print SR6CMDS "REM Could not determine 'item' (partition number) for sector no. $pos_sect.  Check the \$starting[] and \$ending[] arrays.\r\n";
        next;
     }
     $end_sect = $pos_sect + $len_sect - 1;
     if ($end_sect == $pos_sect)  {
        $end_item = $item;
     } else {
        $end_item = &item_from_sect( $end_sect );
        if ($item != $end_item) {
           print SR6CMDS "REM Can't handle sector range $pos_sect-$end_sect because it spans a partition boundary.\r\n";
           # I really need to fix this, to split the range and generate two Spinrite commands.  When I get a round tuit.
           next;
        }
     }

     # Given the starting and ending sector numbers, calculate a SpinRite 6.0
     # "percentage range" which covers that range of sector numbers.  It turns
     # out that SpinRite very slightly miscalculates the sector numbers when
     # converting back from percentages to sector numbers, so our calculations
     # attempt to compensate for that by slightly broadening the "percentage
     # range" (mostly on the lower end):
     $start_of_partition = $starting[$item];
     $size_of_partition = ($ending[$item] + 1 - $starting[$item]);
     $percentage_start = ($pos_sect - $start_of_partition) / $size_of_partition;
     $percentage_end = ($end_sect - $start_of_partition) / $size_of_partition;
     printf SR6CMDS "REM  %1.10f - %1.10f\r\n", $percentage_start, $percentage_end;
     $percentage_start *= 1000000;
     $percentage_start = int($percentage_start * 0.99998891) - 1;  # the "*0.99998891" and "-1" are fudge factors, to (hopefully) compensate for SpinRite's imprecise percentage-to-sector calculations.  This fudge factor *IS* necessary.
     $percentage_start /= 10000;
     $percentage_start += 0.0000000000001;  # avoid effects of Perl 4's rounding errors
     $percentage_end *= 1000000;
     $percentage_end = int($percentage_end + 0.9999999) + 1;  # the "+1" is a fudge factor, because SpinRite's calculations seem to be imprecise.  But this fudge factor might not be necessary.
     $percentage_end /= 10000;
     $percentage_end += 0.0000000000001;

     # truncate to 4 digits past the decimal point, which is the most SpinRite will accept
     $percentage_start = sprintf( '%0.4f', $percentage_start );
     $percentage_end = sprintf( '%0.4f', $percentage_end );

     # Finally, emit a SpinRite command to recover data from that HDD region:
     &emitit( $item, $percentage_start, $percentage_end );

     # Note that emitit() doesn't write the SpinRite command immediately.
     # Instead, it delays writing it, so that it can merge the SpinRite commands
     # for processing adjacent or overlapping regions.  That happens a lot when
     # there are bad disk sectors close together, because the "granularity" with
     # which sectors can be specified using SpinRite 6.0's "percentage offset"
     # scheme is measured in hundreds of sectors, for large partitions.
  }
  &emit_prev();
  return;
}

&doit;

print SR6CMDS "REM  $numlines lines processed.\r\n";
close( SR6CMDS );

print "Created 'SR6CMDS.BAT'\n";

__END__
